/*
 * Transwarp V1.4.3, optimized for direction Amiga -> PC with BaudBandit.device
 *                 38400 bps for PC -> Amiga, uses serial.device
 * Based on Transdisk V4.2
 * Enhancements for the serial port, usage and functionality 1998, 2002, 2004-05 by Wolfgang Stoeggl
 * Copyright 1995-97 Bernd Schmidt, Marcus Sundberg, Stefan Ropke,
 *                   Rodney Hester, Joanne Dow
 *
 * Use DICE and 2.0 includes or above to compile. SAS C works too.
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <strings.h>

#include <exec/devices.h>
#include <exec/io.h>
#include <exec/memory.h>

#include <devices/trackdisk.h>
#include <devices/serial.h>

#include <clib/alib_protos.h>
#include <clib/exec_protos.h>

#include <proto/exec.h>

#define DEBUG 0 /* 0 NODEBUG, 1 DEBUG (serial debugging with DebugIoRequest) */
#define SERIAL_FLAGS SERF_7WIRE|SERF_RAD_BOOGIE

char *prog_name = "TransWarp";   /* Name of program being run */
char *prog_version = "1.4.3"; /* Version of the program */
char *serialdevice = "BaudBandit.device"; /* Default for Amiga -> PC */
/* serial.device selected below for PC -> Amiga or serdevicebuf */
char *inbuffer;
char bad_sec[512];

int baudrate;
int readbuffer = 8192; /* min 4096 for max speed */
int error;
int length;
int blength;
int unit = 0;
int sectors = 11; /* used in main, closeall, openall */
int serial = 0;
int singledrive=0;
BOOL ser_open = FALSE;
BOOL device_open = FALSE;
BOOL buffer_alloc = FALSE;

UBYTE *buffer;
UBYTE *serbuffer;
UBYTE *diskbuffer;

unsigned long crc = ~0;
unsigned long table[256];

struct IOStdReq *ioreq;
struct IOExtSer *ior; /* request block for serial */
struct MsgPort *port;
struct MsgPort *SerialMP;

static void usage(void);

static void usage(void)
{
    fprintf(stderr, "%s: version %s\n", prog_name, prog_version );
    fprintf(stderr, "Usage: transwarp options\n");
    fprintf(stderr, "Recognized options:\n");
    fprintf(stderr, "-b n: Set the baudrate to n\n");
    fprintf(stderr, "-h: Assume device is high density\n");
    fprintf(stderr, "-1: Single drive mode\n");
    fprintf(stderr, "-g: Generate CRC-32\n");
    fprintf(stderr, "-d device unit: Use this device instead of DF0:\n");
    fprintf(stderr, "-c device: Use this serial device instead of BaudBandit or serial\n");
    fprintf(stderr, "-r n: Set the serial read buffer to n\n");
    fprintf(stderr, "-s n: Begin transfer at track n\n");
    fprintf(stderr, "-e n: End transfer at track n\n");
    fprintf(stderr, "-w filename: writes the ADF file <filename> to disk\n\n");
    fprintf(stderr, "Example:\n");
    fprintf(stderr, "transwarp -b 115200 -d trackdisk 1 -s 0 -e 39\n");
    fprintf(stderr, "  transfers the first half of the floppy in DF1: at a\n");
    fprintf(stderr, "  baudrate of 115200 using the serial port.\n");
    fprintf(stderr, "Or:\n");
    fprintf(stderr, "transwarp -w ser: -b 28800 -1\n");
    fprintf(stderr, "  writes the ADF-file from the serial port at a baudrate of 28800\n");
    fprintf(stderr, "  to the disk in df0: after a request for confirmation.\n");
}

/* Declare functions: */
void OpenAll( void );
void CloseAll( void );
void MotorOff( void );
void BadSector( void );
void MakeCRC32(unsigned long * table, unsigned long ID);
void DoCRC32(UBYTE * mem, int size);
void __regargs _CXBRK(void);

/* Possibility to abort TransWarp with CTRL-C and clean up */
/* Do not compile with NoCheckAbort option */
void __regargs _CXBRK(void)
{
    MotorOff();
    CloseAll();
    exit(1);
}

#if DEBUG
void DebugIoRequest (struct IOExtSer * iorequest)
{
    int mask = iorequest->io_SerFlags;

    fprintf(stderr, "CtlChar    0x%lx \n", iorequest->io_CtlChar);
    fprintf(stderr, "RBufLen    %ld   \n", iorequest->io_RBufLen);
    fprintf(stderr, "ExtFlags   0x%lx \n", iorequest->io_ExtFlags);
    fprintf(stderr, "Baud       %ld   \n", iorequest->io_Baud);
    fprintf(stderr, "BrkTime    %ld   \n", iorequest->io_BrkTime);
    fprintf(stderr, "ReadLen    %ld   \n", iorequest->io_ReadLen);
    fprintf(stderr, "WriteLen   %ld   \n", iorequest->io_WriteLen);
    fprintf(stderr, "StopBits   %ld   \n", iorequest->io_StopBits);
    fprintf(stderr, "SerFlags   0x%lx \n", iorequest->io_SerFlags);
    if (mask) {
        fprintf(stderr, "           ");
        if (mask & SERF_XDISABLED)
            fprintf(stderr, "XDISABLED ");
        if (mask & SERF_EOFMODE)
            fprintf(stderr, "EOFMODE ");
        if (mask & SERF_SHARED)
            fprintf(stderr, "SHARED ");
        if (mask & SERF_RAD_BOOGIE)
            fprintf(stderr, "RAD_BOOGIE ");
        if (mask & SERF_QUEUEDBRK)
            fprintf(stderr, "QUEUEDBRK ");
        if (mask & SERF_7WIRE)
            fprintf(stderr, "7WIRE ");
        if (mask & SERF_PARTY_ODD)
            fprintf(stderr, "PARTY_ODD ");
        if (mask & SERF_PARTY_ON)
            fprintf(stderr, "PARTY_ON");
        fprintf(stderr, "\n");
    }
    fprintf(stderr, "Status     0x%lx \n", iorequest->io_Status);
    fprintf(stderr, "\n");

    return;
}
#endif

void
OpenAll( void )
{
    if ((SerialMP = (struct MsgPort *) CreatePort(0,0)) == NULL) {
        fprintf(stderr, "Can't create message port.\n");
        CloseAll();
        exit(30);
    }
    if ((ior = (struct IOExtSer *) CreateExtIO(SerialMP, sizeof(struct IOExtSer))) == NULL) {
        fprintf(stderr, "Can't allocate storage for ExtIO.\n");
        CloseAll();
        exit(30);
    }
    /* only SERF_SHARED and SERF_7WIRE are accepted before opening device */
    ior->io_SerFlags = SERIAL_FLAGS; /* Set before opening device */

    if ((error = OpenDevice(serialdevice,0,(struct IORequest*)ior, 0)) != NULL) {
        fprintf(stderr, "Can't open %s unit %ld, error=%ld\n", serialdevice, (long) unit, error);
        CloseAll();
        exit(30);
    }
    ser_open = TRUE;
#if DEBUG
    DebugIoRequest (ior);
#endif
    if (baudrate != 0) { /* Set baud rate, parameters if needed. */
        ior->IOSer.io_Command = SDCMD_QUERY;
        ior->IOSer.io_Actual = 0;
        if (DoIO((struct IORequest*)ior) != NULL) {
            fprintf(stderr, "Failed SDCMD_QUERY\n");
            CloseAll();
            exit(30);
        }

        if ((length = ior->IOSer.io_Actual) > 0) {
            if ((inbuffer = malloc(length)) == NULL) {
                fprintf(stderr, "Out of memory\n");
                CloseAll();
                exit(30);
            }
            ior->IOSer.io_Command = CMD_READ;
            ior->IOSer.io_Length = length;
            ior->IOSer.io_Data = inbuffer;
            if (DoIO((struct IORequest*)ior) != NULL) {
                fprintf(stderr, "Failed serial read.\n");
                free(inbuffer);
                CloseAll();
                exit(30);
            }
            /* fprintf(stderr,"%s\n",inbuffer); /*If you want to see the nag screen*/
            free(inbuffer);
        }
        ior->io_SerFlags = SERIAL_FLAGS; /* Set again before SDCMD_SETPARAMS */
        ior->io_ExtFlags = 0;
        ior->io_Baud = baudrate;
        ior->io_RBufLen = readbuffer;
        ior->IOSer.io_Command = SDCMD_SETPARAMS;
        if (DoIO((struct IORequest*)ior) != NULL) {
            fprintf(stderr, "Unable to set parameters.\n");
        }
#if DEBUG
        DebugIoRequest (ior);
#endif
    }
}

void
CloseAll( void )
{
/* Clean up serial device */
    if (ser_open) {
        /* Reset the serial device: */
        ior->IOSer.io_Command = CMD_RESET;
        if ((error = DoIO((struct IORequest*)ior)) != NULL)
            fprintf(stderr, "Could not reset the serial device, error=%d\n", error);

        /* Check the io_Error */
        if( ior->IOSer.io_Error )
            fprintf(stderr, "Problems with serial device!\n");

        fprintf(stderr, "Closing %s ...\n", serialdevice);
        CloseDevice((struct IORequest *)ior);
    }
    if (ior) {
        DeleteExtIO((struct IORequest *)ior);
    }
    if (SerialMP) {
        DeletePort(SerialMP);
    }

/* Clean up trackdisk device */
    if (device_open) {
        CloseDevice( (struct IORequest *) ioreq);
        DeleteStdIO(ioreq);
        DeletePort(port);
    }

/* Clean up memory */
    if (buffer_alloc) {
        FreeMem(diskbuffer, sectors*512);
        FreeMem(serbuffer, sectors*512);
        FreeMem(buffer, 512);
    }

/* Finish with a new line */
    fprintf(stderr, "\n");
}

void
MotorOff( void )
{
    ioreq->io_Command = TD_MOTOR;   /* Turn Disk-motor off */
    ioreq->io_Length = 0;
    DoIO( (struct IORequest *) ioreq);
}

void
BadSector( void )
{
    int i;
    char *temp = "unreadablesector"; /* 16 chars */

    for (i = 1; i < 6;) {   /* 5*16 = 512 bytes */
        sprintf(bad_sec, "%s%s", temp, temp);
        temp = bad_sec;
        i++;
    }
#if DEBUG
    fprintf(stderr, "bad_sec created. length: %d\n", strlen(bad_sec));
#endif
}

void MakeCRC32(unsigned long * table, unsigned long ID)
{
    unsigned long i, j, k;

    for(i = 0; i < 256; ++i) {
        k = i;

        for(j = 0; j < 8; ++j) {
            if(k & 1)
                k = (k >> 1) ^ ID;
            else
                k >>= 1;
        }
        table[i] = k;
    }
}

void DoCRC32(UBYTE * mem, int size) {
    while(size--)
        crc = table[(crc ^ *(mem++)) & 0xFF] ^ (crc >> 8);
}


int main(int argc, char *argv[])
{
    char *filename, *openMode="rb";
    FILE *ADFFile;

    int write=0;
    int otherserial=0;
    char devicebuf[256];
    char serdevicebuf[256];
    char *devicename = "trackdisk.device";
    char devicenum = 0;
    int repeat = TRUE;
    int key;
    int ignoreall = FALSE;
    int nocrc = TRUE;            /* Do not calculate CRC by default */
    int i;
    int starttr = 0, endtr = 79; /* int sectors = 11; is now global */

    if (argc == 1) {
        usage();
        exit (1);
    }
    for (i = 1; i < argc;) {
        if (argv[i][0] != '-' || argv[i][2] != 0) {
            usage();
            exit(1);
        }
        switch (argv[i][1]) {
        case 'h':
            sectors = 22;
            i++;
            break;
        case '1':
            singledrive =1;
            i++;
            break;
        case 'g':
            nocrc = FALSE;
            i++;
            break;
        case 'd':
            if (i+2 >= argc) {
                usage();
                exit(1);
            }
            devicenum = atoi(argv[i+2]);
            sprintf(devicebuf, "%s.device", argv[i+1]);
            devicename = devicebuf;
            i += 3;
            break;
        case 'c':
            if (i+1 >= argc) {
                usage();
                exit(1);
            }
            sprintf(serdevicebuf, "%s.device", argv[i+1]);
            otherserial = 1;
            serialdevice = serdevicebuf;
            i += 2;
            break;
        case 's':
            if (i+1 >= argc) {
                usage();
                exit(1);
            }
            starttr = atoi(argv[i+1]);
            i += 2;
            break;
        case 'e':
            if (i+1 >= argc) {
                usage();
                exit(1);
            }
            endtr = atoi(argv[i+1]);
            i += 2;
            break;
        case 'b':
            if (i+1 >= argc) {
                usage();
                exit(1);
            }
            baudrate = atoi(argv[i+1]);
            serial =1;
            i += 2;
            break;
        case 'r':
            if (i+1 >= argc) {
                usage();
                exit(1);
            }
            readbuffer = atoi(argv[i+1]);
            i += 2;
            break;
        case 'w':
            if (i+1 >= argc) {
                usage();
                exit(1);
            }
            filename=argv[i+1];
            write=1;
            i += 2;
            break;
        default:
            usage();
            exit(1);
        }
    }

    if (serial) {
        if (write && !otherserial) {         /* Except serdevicebuf is given */
            serialdevice = "serial.device";  /* Use serial.device for receiving like trans28 */
        }
        fprintf(stderr,"Using %s, baudrate %d, buffer %d\n", serialdevice, baudrate, readbuffer);
    }
    fprintf(stderr,"Using %s unit %d\n", devicename, devicenum);
    fprintf(stderr,"Tracks are %d sectors\n", sectors);
    fprintf(stderr,"First track %d, last track %d\n", starttr, endtr);

    buffer = AllocMem(512, MEMF_CHIP);
    serbuffer = AllocMem(sectors*512, MEMF_CHIP);
    diskbuffer = AllocMem(sectors*512, MEMF_CHIP);
    buffer_alloc = TRUE;
    if (nocrc == FALSE)
        MakeCRC32(table, 0xEDB88320); /* Create the CRC32 table */

    if (serial)
        OpenAll();
    if (write) {
        if (!serial) {
            ADFFile = fopen(filename,openMode);

            if (!ADFFile) {
                fprintf(stderr,"Error while opening input file\n");
                exit (1);
            }
        }
    }
    while(repeat) {
        crc = ~0; /* Set CRC back to ~0 for each disk */
        if (singledrive) {
            fprintf(stderr,"Insert disk and hit RETURN to continue (Q to quit):\n");
            fflush(stderr);
            key = getch();
            if ((key == 'q')||(key == 'Q')) {
                CloseAll();
                exit(1);
            }
        }

        port = CreatePort(0, 0);
        if (port) {
            ioreq = CreateStdIO(port);
            if (ioreq) {
                if (OpenDevice(devicename, devicenum, (struct IORequest *) ioreq, 0) == 0) {
                    int tr, sec;
                    device_open = TRUE;
                    if (write) {
                        do {
                            /* Check if the disk is write protected or not: */
                            ioreq->io_Command = TD_PROTSTATUS;
                            DoIO( (struct IORequest *) ioreq );
                            /* Is the disk write protected or not? */
                            if( ioreq->io_Actual ){
                                fprintf(stderr,"Remove write protection and hit any key to continue!\n" );
                                fflush(stderr);
                                getch();
                            }
                            else
                                serial ? fprintf(stderr,"Send the data NOW!\n" ) : fprintf(stderr,"OK\n" );
                        } while (ioreq->io_Actual); /* Repeat until disk is not write protected */
                    }

                    ioreq->io_Command = TD_SEEK;                       /* Position the head to start track */
                    ioreq->io_Offset = 512 * (starttr * 2 * sectors);  /* Move to position */
                    DoIO( (struct IORequest *) ioreq );                /* Do request: */

                    ioreq->io_Command = write ? CMD_WRITE : CMD_READ;
                    ioreq->io_Length = 512;
                    ioreq->io_Data = buffer;
                    for (tr = starttr*2; tr < (endtr+1)*2; tr++) {
                        fprintf(stderr,"Track: %d\r",tr/2);
                        for (sec = 0; sec < sectors; sec++) {
                            fflush(stderr);
                            if (write) {
                                if (serial) {                      /* begin simple serial reading */
                                    ior->IOSer.io_Command = CMD_READ;
                                    ior->IOSer.io_Length = 512;
                                    ior->IOSer.io_Data = buffer;
                                    if ((error = DoIO((struct IORequest*)ior)) != NULL) {
                                        fprintf(stderr, "\nFailed serial read, error=%d\n", error);
                                        MotorOff();
                                        CloseAll();
                                        exit(30);
                                    }
                                }                                  /* end simple serial reading */
                                if (!serial) {
                                    if (fread(buffer, sizeof(UBYTE), 512, ADFFile) < 512) {
                                        fprintf(stderr, "\nError: ADF file too short?\n");
                                        MotorOff();
                                        CloseAll();
                                        exit(1);
                                    }
                                }
                            }
                            ioreq->io_Offset = 512 * (tr * sectors + sec);
                            do {
                                if (error = DoIO( (struct IORequest *) ioreq)) {
                                    fprintf(stderr, "Disk error %d (track %d, head %d, sector %d). Retry (y/n/i/a/q)? ", error, tr/2, tr%2, sec);
                                    fflush(stderr);
                                    if (!ignoreall)
                                        key = getch();
                                    if (key == 'a') {
                                        fprintf(stderr, "all\nAll errors will be ignored.\n");
                                        fprintf(stderr, "Select n or i for further error handling! ");
                                        ignoreall = TRUE;
                                        fflush(stderr);
                                        if ((key = getch()) != 'n')
                                            key = 'i';
                                    }

                                    switch (key) {
                                    case 'n':
                                        fprintf(stderr, "no\n");
                                        if (!write) {
                                            if (!strlen(bad_sec)) {
                                                BadSector();
                                            }
                                            memcpy (buffer, bad_sec, 512); /* otherwise the previous sector is still in buffer */
                                            fprintf(stderr, "Sector filled with 'unreadablesector'\n");
                                        }
                                        else
                                            fprintf(stderr, "Use error free formatted disks!\n");
                                        error = NULL;
                                        break;
                                    case 'i':
                                        fprintf(stderr, "ignore\n");
                                        if (!write)
                                            fprintf(stderr, "Warning - error ignored!\n");
                                        else
                                            fprintf(stderr, "Use error free formatted disks!\n");
                                        error = NULL;
                                        break;
                                    case 'q':
                                        fprintf(stderr, "quit\n");
                                        MotorOff();
                                        CloseAll();
                                        exit(1);
                                    default:
                                        fprintf(stderr, "yes\n");
                                        break;
                                    }
                                }
                            } while (error); /* Try until there is no error or it is ignored */


                            if (!write) {
                                if (serial) {
                                    memcpy (diskbuffer+sec*512, buffer, 512); /* Buffering sectors */
                                }
                                if (!serial)
                                    fwrite(buffer, sizeof(UBYTE), 512, stdout);
                            }

                            if (nocrc == FALSE)
                                DoCRC32(buffer, 512); /* Calculate CRC for each sector */
                        }
                        if (serial && !write) {
                            WaitIO((struct IORequest*)ior); /* Wait until previous request is finished */
                            memcpy (serbuffer, diskbuffer, sectors*512); /* Copy whole track to serbuffer */
                            ior->IOSer.io_Command = CMD_WRITE;
                            ior->IOSer.io_Length = sectors*512;
                            ior->IOSer.io_Data = serbuffer;
                            /* fprintf(stderr, "\nwaited\n"); */
                            SendIO((struct IORequest *)ior); /* Continue reading from disk in the meantime */
                        }
                    }

                    if (write) {        /* Make sure the last track is written to disk */
                        ioreq->io_Command = CMD_UPDATE;
                        DoIO( (struct IORequest *) ioreq);
                    }
                    MotorOff();
                    CloseDevice( (struct IORequest *) ioreq);
                    device_open = FALSE;
                    ignoreall = FALSE;
                } else
                    fprintf(stderr,"Unable to open %s unit %d\n", devicename, devicenum);
                DeleteStdIO(ioreq);
            }
            DeletePort(port);
        }
        if (nocrc == FALSE)
            fprintf(stderr,"\nCRC32: %08lX", ~crc);
        if (serial) { /* Possibility to repeat serial transfers */
            fprintf(stderr,"\n\nHit RETURN for the next disk or Q to quit:\n");
            fflush(stderr);
            key = getch();
            if ((key == 'q')||(key == 'Q')) {
                repeat = FALSE;
            }
            singledrive = FALSE; /* Otherwise 2x similar dialog */
        }
        else
            repeat = FALSE;
    }
    if (write) {
        if (!serial)
            fclose (ADFFile);
    }
    if (serial) {
        WaitIO((struct IORequest *)ior); /* Wait for serial IO to finish at the end */

    }
    CloseAll();
    return 0;
}
